@@ -24,6 +24,14 @@
 #ifdef HAVE_STDIO_H
 #include <stdio.h>
 #endif
+#ifdef HAVE_GETPEERNAME
+#ifdef HAVE_SYS_SOCKET_H
+#include <sys/socket.h>
+#endif
+#ifdef HAVE_NETINET_IN_H
+#include <netinet/in.h>
+#endif
+#endif
 
 #include "net.h"
 #include "opennet.h"
@@ -39,6 +47,26 @@ int opennet_init(void) {
 	return(0);
 }
 
+/*
+	Replace 
+		@@OSNM@@	OS Name (linux, freebsd, sunos, etc)
+		@@OSVR@@	OS version (2.2.x, 4.2, 5.8, etc)
+		@@OSVS@@	OS version (short) (2.2, 4.2, 5.8, etc)
+		@@ARCH@@	Arch (i386, sparc64, sun4u, sun4m, etc)
+*/
+char *parse_url_subst(const char *src) {
+	char *retval = NULL;
+
+	if (!strstr(src, "@@")) {
+		return(strdup(src));
+	}
+
+	return(retval);
+}
+
+/*
+ * Parse a URL into its components
+ */
 static int parse_url(char *url, char **scheme, char **user, char **pass, char **host, int *port, char **path) {
 	char *tmptr, *schemeptr, *pathptr, *userptr, *passptr, *hostptr, *portptr;
 
@@ -109,9 +137,62 @@ static int parse_url(char *url, char **scheme, char **user, char **pass, char **
 	return(0);
 }
 
-static int open_net_http(char *user, char *pass, char *host, int port, char *path, int flags) {
+static ssize_t read_net_internal(int fd, void *buf, size_t count, int socketStatus) {
+	ssize_t retval = 0;
+	ssize_t read_ret;
+	int socket = 0;
+#ifdef HAVE_GETPEERNAME
+	int gpn_ret;
+	struct sockaddr_in tmp;
+	int tmplen;
+
+	if (socketStatus < 0) {
+		tmplen = sizeof(tmp);
+
+		gpn_ret = getpeername(fd, (struct sockaddr *) &tmp, &tmplen);
+
+		if (gpn_ret >= 0) {
+			socket = 1;
+		} else {
+			socket = 0;
+		}
+	} else {
+		socket = socketStatus;
+	}
+#else
+	socket = socketStatus;
+#endif
+
+	while (count) {
+		if (socket) {
+			read_ret = recv(fd, buf, count, 0);
+		} else {
+			read_ret = read(fd, buf, count);
+		}
+
+		if (read_ret == 0) {
+			break;
+		}
+
+		if (read_ret < 0) {
+			retval = read_ret;
+			break;
+		}
+
+		count -= read_ret;
+		buf += read_ret;
+		retval += read_ret;
+	}
+
+	return(retval);
+}
+
+static int open_net_http(char *user, char *pass, char *host, int port, char *path, int flags, off_t offset) {
 	ssize_t send_ret, recv_ret;
 	char httpcmd[4096], charbuf, httpreply[1024], *http_response;
+	char garbage[4096];
+	ssize_t read_ret, sink_bytes = 0;
+	int bytes_to_read;
 	int fd;
 	int snprintf_ret;
 	int http_response_code = 0;
@@ -123,7 +204,11 @@ static int open_net_http(char *user, char *pass, char *host, int port, char *pat
 		return(-1);
 	}
 
-	snprintf_ret = snprintf(httpcmd, sizeof(httpcmd), "GET /%s HTTP/1.0\015\012Host: %s\015\012\015\012", path, host);
+	if (offset == 0) {
+		snprintf_ret = snprintf(httpcmd, sizeof(httpcmd), "GET /%s HTTP/1.0\015\012Host: %s\015\012\015\012", path, host);
+	} else {
+		snprintf_ret = snprintf(httpcmd, sizeof(httpcmd), "GET /%s HTTP/1.0\015\012Host: %s\015\012Range: bytes=%lu-\015\012\015\012", path, host, (unsigned long) offset);
+	}
 
 	if (snprintf_ret >= sizeof(httpcmd)) {
 		return(-1);
@@ -151,7 +236,18 @@ static int open_net_http(char *user, char *pass, char *host, int port, char *pat
 					http_response = httpreply + 9;
 					http_response_code = strtoul(http_response, NULL, 10);
 					switch (http_response_code) {
+						case 206:
+							if (offset == 0) {
+								CHECKPOINT;
+								net_close(fd);
+								return(-1);
+							}
+							break;
 						case 200:
+							if (offset != 0) {
+								sink_bytes = offset;
+							}
+							break;
 						case 301:
 						case 302:
 							break;
@@ -178,7 +274,6 @@ static int open_net_http(char *user, char *pass, char *host, int port, char *pat
 			httpreply[i++] = charbuf;
 		}
 
-
 		if (newlinecount == 4) {
 			break;
 		}
@@ -189,30 +284,68 @@ static int open_net_http(char *user, char *pass, char *host, int port, char *pat
 		}
 	}
 
+	if (sink_bytes) {
+		while (offset) {
+			bytes_to_read = offset;
+			if (bytes_to_read > sizeof(garbage)) {
+				bytes_to_read = sizeof(garbage);
+			}
+			SPOTVAR_LU(bytes_to_read);
+
+			read_ret = read_net_internal(fd, garbage, bytes_to_read, 1);
+			if (read_ret <= 0) {
+				SPOTVAR_I(read_ret);
+				net_close(fd);
+				return(-1);
+			}
+
+			offset -= read_ret;
+		}
+	}
+
 	return(fd);
 }
 
-static int open_net_ftp(char *user, char *pass, char *host, int port, char *path, int flags) {
+static int open_net_ftp(char *user, char *pass, char *host, int port, char *path, int flags, off_t offset) {
 	return(-1);
 }
 
-static int open_net_internal(const char *pathname, int flags, mode_t mode, int *socket) {
+static int open_net_internal(const char *pathname, int flags, mode_t mode, int *socket, off_t offset) {
 	char *pathname_dup;
 	char *scheme, *user, *pass, *host, *path;
+	off_t lseek_ret;
 	int port = 0;
 	int parse_ret;
 	int retval = -1;
 
-	*socket = 0;
+	if (socket) {
+		*socket = 0;
+	}
 
 	if (!pathname) {
-		return(open(pathname, flags, mode));
+		retval = open(pathname, flags, mode);
+		if (offset && retval >= 0) {
+			lseek_ret = lseek(retval, offset, SEEK_SET);
+			if (lseek_ret != offset) {
+				close(retval);
+				return(-1);
+			}
+		}
+		return(retval);
 	}
 
 	pathname_dup = strdup(pathname);
 
 	if (!pathname_dup) {
-		return(open(pathname, flags, mode));
+		retval = open(pathname, flags, mode);
+		if (offset && retval >= 0) {
+			lseek_ret = lseek(retval, offset, SEEK_SET);
+			if (lseek_ret != offset) {
+				close(retval);
+				return(-1);
+			}
+		}
+		return(retval);
 	}
 
 	parse_ret = parse_url(pathname_dup, &scheme, &user, &pass, &host, &port, &path);
@@ -223,19 +356,28 @@ static int open_net_internal(const char *pathname, int flags, mode_t mode, int *
 	}
 
 	if (strcasecmp(scheme, "http") == 0) {
-		retval = open_net_http(user, pass, host, port, path, flags);
+		retval = open_net_http(user, pass, host, port, path, flags, offset);
 	}
 
 	if (strcasecmp(scheme, "ftp") == 0) {
-		retval = open_net_ftp(user, pass, host, port, path, flags);
+		retval = open_net_ftp(user, pass, host, port, path, flags, offset);
 	}
 
 	free(pathname_dup);
 
 	if (retval < 0) {
 		retval = open(pathname, flags, mode);
+		if (offset && retval >= 0) {
+			lseek_ret = lseek(retval, offset, SEEK_SET);
+			if (lseek_ret != offset) {
+				close(retval);
+				return(-1);
+			}
+		}
 	} else {
-		*socket = 1;
+		if (socket) {
+			*socket = 1;
+		}
 	}
 
 	return(retval);
@@ -251,7 +393,7 @@ NETFILE *fopen_net(const char *pathname, const char *mode) {
 		return(NULL);
 	}
 
-	ret->fd = open_net_internal(pathname, O_RDONLY, 0666, &is_sock);
+	ret->fd = open_net_internal(pathname, O_RDONLY, 0666, &is_sock, 0);
 
 	if (ret->fd < 0) {
 		free(ret);
@@ -263,14 +405,9 @@ NETFILE *fopen_net(const char *pathname, const char *mode) {
 	ret->bufused = 0;
 	ret->eof = 0;
 	ret->socket = is_sock;
-
-	if (is_sock) {
-//		ret->recv_func = recv;
-//		ret->send_func = send;
-	} else {
-//		ret->recv_func = read;
-//		ret->send_func = write;
-	}
+	ret->pos = 0;
+	ret->length = -1;
+	ret->url = strdup(pathname);
 
 	return(ret);
 }
@@ -321,6 +458,10 @@ size_t fread_net(void *ptr, size_t size, size_t nmemb, NETFILE *stream) {
 
 	if (stream->bufused < bytes_to_copy) {
 		bytes_to_copy = stream->bufused;
+
+		/* Adjust bytes_to_copy to be a multiple of size */
+		bytes_to_copy /= size;
+		bytes_to_copy *= size;
 	}
 
 	memcpy(ptr, stream->buf, bytes_to_copy);
@@ -328,12 +469,14 @@ size_t fread_net(void *ptr, size_t size, size_t nmemb, NETFILE *stream) {
 	stream->bufused -= bytes_to_copy;
 	stream->bufsize -= bytes_to_copy;
 
-	if (stream->bufused == 0) {
+	if ((stream->bufused < size && size < (stream->bufsize_s / 2)) || stream->bufused == 0) {
+		memmove(stream->buf_s, stream->buf, stream->bufused);
 		stream->buf = stream->buf_s;
 		stream->bufsize = stream->bufsize_s;
 	}
 
 	retval = bytes_to_copy / size;
+	stream->pos += retval;
 
 	return(retval);
 }
@@ -401,6 +544,7 @@ char *fgets_net(char *s, int size, NETFILE *stream) {
 	stream->buf += bytes_to_copy;
 	stream->bufused -= bytes_to_copy;
 	stream->bufsize -= bytes_to_copy;
+	stream->pos += bytes_to_copy;
 
 	if (stream->bufused == 0) {
 		stream->buf = stream->buf_s;
@@ -425,6 +569,9 @@ int fclose_net(NETFILE *stream) {
 	if (stream->buf_s) {
 		free(stream->buf_s);
 	}
+	if (stream->url) {
+		free(stream->url);
+	}
 
 	free(stream);
 
@@ -440,49 +587,11 @@ int fclose_net(NETFILE *stream) {
 int open_net(const char *pathname, int flags, mode_t mode) {
 	int tmp;
 
-	return(open_net_internal(pathname, flags, mode, &tmp));
+	return(open_net_internal(pathname, flags, mode, &tmp, 0));
 }
 
-
 ssize_t read_net(int fd, void *buf, size_t count) {
-	ssize_t retval = -1;
-	ssize_t read_ret;
-	int socket = 0;
-#ifdef HAVE_GETPEERNAME
-	int gpn_ret;
-	struct sockaddr_in *tmp = NULL;
-	int tmplen;
-
-	tmplen = sizeof(tmp);
-
-	gpn_ret = getpeername(fd, (struct sockaddr *) tmp, &tmplen);
-
-	if (gpn_ret >= 0) {
-		socket = 1;
-	}
-#endif
-
-	while (count) {
-		if (socket) {
-			read_ret = recv(fd, buf, count, 0);
-		} else {
-			read_ret = read(fd, buf, count);
-		}
-
-		if (read_ret == 0) {
-			break;
-		}
-
-		if (read_ret < 0) {
-			retval = read_ret;
-			break;
-		}
-
-		count -= read_ret;
-		buf += read_ret;
-	}
-
-	return(retval);
+	return(read_net_internal(fd, buf, count, -1));
 }
 
 off_t lseek_net(int filedes, off_t offset, int whence) {
@@ -490,5 +599,72 @@ off_t lseek_net(int filedes, off_t offset, int whence) {
 }
 
 int fseek_net(NETFILE *stream, long offset, int whence) {
-	return(-1);
+	static char garbage[4096];
+	size_t fread_ret, bytes_to_read;
+	off_t newoffset, lseek_ret;
+	int newfd;
+
+	switch (whence) {
+		case SEEK_CUR:
+			newoffset = stream->pos + offset;
+			break;
+		case SEEK_END:
+			newoffset = stream->length - offset;
+			break;
+		case SEEK_SET:
+			newoffset = offset;
+			break;
+		default:
+			CHECKPOINT;
+			return(-1);
+			break;
+	}
+
+	lseek_ret = lseek(stream->fd, newoffset, SEEK_SET);
+	if (lseek_ret != ((off_t) -1)) {
+		stream->pos = lseek_ret;
+		stream->buf = stream->buf_s;
+		stream->bufused = 0;
+		stream->bufsize = stream->bufsize_s;
+
+		return(0);
+	}
+
+	/*
+	 * Read up to 4K of data to avoid opening a new connection
+	 */
+	if (newoffset >= stream->pos && (newoffset - stream->pos < sizeof(garbage))) {
+		bytes_to_read = newoffset - stream->pos;
+		fread_ret = fread_net(garbage, sizeof(garbage[0]), bytes_to_read, stream);
+
+		if (fread_ret == bytes_to_read) {
+			return(0);
+		}
+	}
+
+	newfd = open_net_internal(stream->url, O_RDONLY, 0666, NULL, newoffset);
+
+	if (newfd < 0) {
+		return(-1);
+	}
+
+	net_close(stream->fd);
+	stream->fd = newfd;
+	stream->pos = newoffset;
+	stream->buf = stream->buf_s;
+	stream->bufused = 0;
+	stream->bufsize = stream->bufsize_s;
+
+	return(0);
+}
+
+long ftell_net(NETFILE *stream) {
+	if (!stream) {
+		return(-1);
+	}
+	if (stream->fd < 0) {
+		return(-1);
+	}
+
+	return(stream->pos);
 }
