@@ -10,13 +10,15 @@ int uname(struct utsname *buf) {
 	SYSTEM_INFO winsysinfo;
 #endif
 
-	if (buf==NULL) return(-1);
+	if (!buf) {
+		return(-1);
+	}
 
 #ifdef _USE_WIN32_
 	GetSystemInfo(&winsysinfo);
-	winver=GetVersion();
+	winver = GetVersion();
                         
-	snprintf(buf->release, 64, "%i.%i", winver&0xff, (winver&0xff00)>>8);
+	snprintf(buf->release, 64, "%i.%i", winver & 0xff, (winver & 0xff00) >> 8);
 	snprintf(buf->machine, 64, "%lu", (unsigned long) winsysinfo.dwProcessorType); 
 	strcpy(buf->sysname, "windows");
 #else
